// diskscan.cc

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>

//3) fix errors at end of disk

int main(int argc,char*argv[])
{
  if ((argc < 2)||(argc > 3))
    {
      fprintf(stderr,"Usage: %s /dev/hdXXX [offsetMB]\n",argv[0]);
      exit(1);
    }

  setbuf(stdout,0);
  setbuf(stderr,0);

  int tty = isatty(1);

  char*diskname = argv[1];
  int  offsetMB = 0;

  if (argc == 3)
    offsetMB = atoi(argv[2]);

  int rfd = open(diskname,O_RDONLY|O_LARGEFILE);
  if (rfd < 0)
    {
      fprintf(stderr,"Cannot open file \'%s\', errno: %d (%s)\n",diskname,errno,strerror(errno));
      return -1;
    }

  int countSlow = 0;
  int countUnreadable = 0;
  int countShort = 0;
  double longestReadTime = 0;
  double maxRate = -HUGE_VAL;
  double minRate = +HUGE_VAL;

  int incr = 10;
  int position = offsetMB;

  off64_t mb = 1024*1024;

  int lastShortRead = 0;

  int eof = 0;
  while (!eof)
    {
      off64_t seek = position * mb;

      off64_t status = lseek64(rfd,seek,SEEK_SET);
      if (status < 0)
        {
          fprintf(stderr,"At %d MiBytes: Cannot seek file \'%s\' to %d MiBytes (%lld), errno: %d (%s)\n",position,diskname,position,seek,errno,strerror(errno));
          break;
        }

      double countbytes = 0;
      double counttime  = 0;

      double maxreadtime = 0;

      int shouldRead = incr*mb;
      int didRead = 0;

      while (shouldRead > 0)
        {
          struct timeval tstart, tend;
          int bufsize = 256*1024;
          char buf[bufsize];

          errno = 0;

          int toread = shouldRead;

          if (toread > bufsize)
            toread = bufsize;

          gettimeofday(&tstart,0);
          
          int rd = read(rfd,buf,toread);
          
          gettimeofday(&tend,0);
          
          long diffs = tend.tv_sec  - tstart.tv_sec;
          long diffu = tend.tv_usec - tstart.tv_usec;
          double readtime = diffs + 0.000001 * diffu;

          if (readtime > longestReadTime)
            longestReadTime = readtime;

          if (readtime > maxreadtime)
            maxreadtime = readtime;
          
          if (rd == 0)
            {
              // at EOF, the previous read may have been short, do not count it
              if (lastShortRead)
                countShort--;

              fprintf(stderr,"At %d MiBytes: EOF\n",position);
              eof = 1;
              break;
            }
          else if (rd < 0)
            {
              if (errno != EIO)
                {
                  fprintf(stderr,"At %d MiBytes: Unreadable area at offset %d: Unexpected read error, errno: %d (%s), giving up.\n",position,didRead,errno,strerror(errno));
                  countUnreadable++;
                  eof = 1;
                  break;
                }

              fprintf(stderr,"At %d MiBytes: Unreadable area at offset %d: I/O error, errno: %d (%s)\n",position,didRead,errno,strerror(errno));
              countUnreadable++;

              if (readtime > 200.0)
                {
                  fprintf(stderr,"At %d MiBytes: Extremely slow read at offset %d: %.3f sec, giving up.\n",position,didRead,readtime);
                  countSlow++;
                  eof = 1;
                }

              break;
            }

          shouldRead -= rd;
          didRead    += rd;

          lastShortRead = 0;

          if (rd != toread)
            {
              fprintf(stderr,"At %d MiBytes: Short read at offset %d\n",position,didRead);
              countShort++;
              lastShortRead = 1;
            }
          
          if (readtime > 30.0)
            {
              fprintf(stderr,"At %d MiBytes: Extremely slow read at offset %d: %.3f sec, giving up.\n",position,didRead,readtime);
              countSlow++;
              eof = 1;
              break;
            }

          countbytes += rd;
          counttime  += readtime;

#if 0
          printf("%d %f\n",rd,readtime);
          if (i>100)
            {
              stop = 1;
              break;
            }
#endif
        }

      double rate = countbytes/counttime;

      if (rate > maxRate)
        maxRate = rate;

      if (rate < minRate)
        minRate = rate;

      if (rate < 1*1024*1024 || maxreadtime > 0.5)
        {
          fprintf(stderr,"At %d MiBytes: Slow read around offset %.0f, max read time: %.3f sec, data rate: %.0f MiBytes/sec\n",position,countbytes,maxreadtime,rate/mb);
          countSlow++;
        }

      if (tty)
        printf("At %d MiBytes: %.0f bytes, max read time: %.3f sec, data rate: %.0f MiBytes/sec\r",position,countbytes,maxreadtime,rate/(1024*1024));

      position += incr;
    }

  fprintf(stderr,"Summary: disk size %d GiBytes, %d slow reads, %d short reads, %d unreadable areas, data rate %.0f to %.0f MiBytes/sec, longest read time %.3f sec\n",(position+1023)/1024,countSlow,countShort,countUnreadable,minRate/mb,maxRate/mb,longestReadTime);

  return 0;
}

// end file
