#!/usr/bin/perl -w

# identify all disks

my @disks;
my %hdparm;

# compile disk scanner

#system "g++ -o /tmp/diskscan -g -O2 -Wall -Wuninitialized diskscan.cc";

my $startdate = `date`; chop $startdate;

# look for PATA disks

foreach my $disk ("/dev/hda", "/dev/hdb", "/dev/hdc", "/dev/hdd")
  {
    next if ! -r $disk;
    my $hdparm = `/sbin/hdparm $disk 2>&1`;
    next if ($hdparm =~ /Invalid/);
    next if ($hdparm =~ /No such device/);
    #print "hdparm : $hdparm\n";
    push @disks,$disk;
    $hdparm{$disk} = $hdparm;
  }

# look for SCSI and SATA disks

for (my $sd = "a"; ; $sd++)
  {
    my $disk = "/dev/sd$sd";
    last if ! -r $disk;
    my $hdparm = `/sbin/hdparm $disk 2>&1`;
    last if ($hdparm =~ /Invalid/);
    last if ($hdparm =~ /No such device/);
    #print "hdparm : $hdparm\n";
    push @disks,$disk;
    $hdparm{$disk} = $hdparm;
  }

print "Found disks: ",join(" ",@disks),"\n";

#die "Here!";

# run diskscan on all disks

foreach my $disk (@disks)
  {
    my $xdisk = $disk;
    $xdisk =~ s/\//\_/g;
    my $log = "/var/log/diskscrub$xdisk.txt";

    append($log,"Date: ".`date`);

    # capture smart data

    my $smart1 = `/usr/sbin/smartctl -a $disk`;
    append($log,"\nSMART data before scrubbing:\n\n".$smart1);

    #die "Here!";

    # run the disk scanner

    my $scrublog = "/var/log/diskscrub$xdisk.last.txt";

    system "/usr/sbin/diskscrub.exe $disk >& $scrublog";

    my $scrub = `grep Summary $scrublog`; chop $scrub;
    $scrub =~ /(Summary.*)/;
    $scrub = $1;

    #print "Scrub: $scrub\n";

    append($log,"\nScrubbing summary:\n\n".$scrub);

    # capture smart data

    my $smart2 = `/usr/sbin/smartctl -a $disk`;
    append($log,"\nSMART data after scrubbing:\n\n".$smart2);

    # do a diff between the before and after smart data

    use File::Temp qw/ tempfile /;
    my ($tmpfh1, $tmpfilename1) = tempfile();
    my ($tmpfh2, $tmpfilename2) = tempfile();

    print $tmpfh1 $smart1; close $tmpfh1;
    print $tmpfh2 $smart2; close $tmpfh2;

    my $smartdiff = `diff -u $tmpfilename1 $tmpfilename2`;

    unlink $tmpfilename1;
    unlink $tmpfilename2;

    $result{$disk}{"summary"}   = $scrub;
    $result{$disk}{"scan"}      = `cat $scrublog`;
    $result{$disk}{"smart"}     = $smart2;
    $result{$disk}{"smartdiff"} = $smartdiff;
  }

# report results

my $report = "";

$report .= "\n";
$report .= "Date: $startdate\n";
$report .= "Host: ".`hostname`;
$report .= "\n";
$report .= ("="x79)."\n";
$report .= "                  Size    Slow   Short    Read  Minrate  Maxrate  Slowest read\n";
$report .= "               GiBytes   reads   reads  errors       MiBytes/sec        sec\n";
$report .= ("-"x79)."\n";
foreach my $disk (@disks)
  {
    my $summary = $result{$disk}{"summary"};
    my @summary = split(/\s+/,$summary);
    my $size = $summary[3];
    my $slowr = $summary[5];
    my $shortr = $summary[8];
    my $unr = $summary[11];
    my $minrate = $summary[16];
    my $maxrate = $summary[18];
    my $slowt = $summary[23];
    #print "xsummary: ",join("|",@summary),"\n";
    my $zsummary = sprintf("%10s: %10d     %3d     %3d     %3d %8s %8s %10.3f",$disk,$size,$slowr,$shortr,$unr,$minrate,$maxrate,$slowt);
    #print "$zsummary\n";
    $report .= $zsummary;
    $report .= "\n";
  }

$report .= ("="x79)."\n";
$report .= "\n";

foreach my $disk (@disks)
  {
    $report .= "\n";
    $report .= "Scan of $disk:\n";
    $report .= "\n";
    $report .= ("="x79)."\n";
    $report .= $result{$disk}{"scan"};
    $report .= ("="x79)."\n";
    $report .= "\n";
    if (length($result{$disk}{"smartdiff"}) > 0)
      {
        $report .= "SMART before and after the scan of $disk:\n";
        $report .= "\n";
        $report .= $result{$disk}{"smartdiff"};
        $report .= "\n";
        $report .= ("="x79)."\n";
      }
  }

$report .= "\n";

foreach my $disk (@disks)
  {
    $report .= "\n";
    $report .= "SMART report for $disk:\n";
    $report .= "\n";
    $report .= $result{$disk}{"smart"};
    $report .= "\n";
  }

my $enddate = `date`; chop $enddate;

$report .= "Date: $enddate\n";
$report .= "\n";

print $report;

exit 0;

sub append
  {
    my $filename = shift @_;
    my $text = join("",@_);
    open(TMPOUT,">>$filename") || die "Cannot write to $filename, error $!";
    print TMPOUT $text;
    close TMPOUT;
  }

#end file
